// Jean-Baptiste PERIN 2021
// osdk_build.bat && osdk_execute.bat
// gcc -D USE_NCURSES -g main.c -lncursesw -o main.exe && main.exe
// gcc -D DEBUG -g main.c -o main.exe && main.exe
//

#define SCREEN_WIDTH                    40
#define SCREEN_HEIGHT                   26

#define START_POS_X 0  
#define START_POS_Y 32 
#define START_POS_Z 8
#define START_ROT_Z -64
#define START_ROT_X 0

// #define DEBUG
#define USE_AUTOPILOT

#include "config.h"

#ifdef USE_MUSIC
extern void StartMusic();
extern void EndMusic();
extern unsigned int MusicResetCounter;
#endif // USE_MUSIC

void init();
void scene1();
void end();


char message [50];




void engLoop()
{
    engRunning = 1;
    while (engRunning == 1)
    {
        key1 = key();
        player();
        // cls();

#ifdef USE_GLORIC
        // empty buffer
        glInitScreenBuffers();
#endif

#ifdef USE_RAYCASTING
        rayInitCasting();
#endif
        engPulse();

#ifdef USE_RAYCASTING
        rayProcessPoints();
#ifdef DEBUG
        detailPoints();
#endif
        rayProcessWalls();
#ifdef DEBUG
    textZBuffer ();
    get();
#endif        
#endif

#ifdef USE_RAYCASTING
#ifdef USE_GLORIC
        walls2Zbuffer ();
#endif
#endif


#ifdef USE_GLORIC
        // project 3D points to 2D coordinates
        glProjectArrays();
        // draw game scene's shapes in buffer
        glDrawFaces();
        glDrawSegments();
        glDrawParticles();

        // update display with buffer
		// sprintf(fbuffer+40*23+10, "%d.%d   ", kernel_s, kernel_cs);
		// sprintf(fbuffer+40*23+20, "%d.%d   ",kernel_beat,  kernel_fraction);
		// sprintf(fbuffer+40*23+30, "%d    ",getTempo());



        //sprintf(fbuffer, " X %d Y %d Z %d a %d", glCamPosX, glCamPosY, glCamPosZ, glCamRotZ);
        glBuffer2Screen();

		// sprintf(message, "%d  ", ii);
		// AdvancedPrint(2,0,message);
		sprintf(message, "%d.%d   ", kernel_s, kernel_cs);
		AdvancedPrint(10,12,message);
		sprintf(message, "%d.%d   ",kernel_beat,  kernel_fraction);
		AdvancedPrint(20,12,message);
		sprintf(message, "%d    ",getTempo());
		AdvancedPrint(30,12,message);

        if (kernel_beat >= 140) engRunning = 0;
#endif

        system();
#ifdef USE_NCURSES
        refresh();
#endif
        if ((key1 == 'q') || (key1 == 'Q')) engRunning = 0;
    }
}

void scene1()
{

#ifdef USE_GLORIC
    glSceneInit();
#endif 

#ifdef USE_RAYCASTING
    raySceneInit (scene_00);
#endif

    engInitObjects();
    
    engAddObject(OBJ_EMPTY, 10, 10, "11 >(>");
    engAddObject(OBJ_FISH, 20, 20, "22 >(>");
    engAddObject(OBJ_CAMERA, START_POS_X, START_POS_Y, "11 >(>");

    // printText("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~", 0, 26);
#ifdef USE_MUSIC
    StartMusic();
#endif // USE_MUSIC
    engLoop();
#ifdef USE_MUSIC
    EndMusic();
#endif // USE_MUSIC

}

void main()
{

    init();
    scene1();
    end();
}

void init(){

#ifdef USE_NCURSES
    int ch;
    initscr();			/* Start curses mode 		*/
	raw();				/* Line buffering disabled	*/
	keypad(stdscr, TRUE);		/* We get F1, F2 etc..		*/
	cbreak();			/* Don't echo() while we do getch */
	noecho();			/* Don't echo() while we do getch */
    nodelay(stdscr, TRUE);
    scrollok(stdscr, TRUE);

	refresh();			/* Print it on to the real screen */
    getch();			/* Wait for user input */

#endif

    cls();
    paper(0);
    ink(7);
    poke(0x26A, CURSOR_START);


}

void end(){

#ifdef USE_NCURSES
   	endwin();			/* End curses mode		  */
#endif
    text();
    cls();
    poke(0x26a, CURSOR_END);
}


#ifdef DEBUG
void detailPoints(){
    unsigned char idxPoint;

    for (idxPoint = 0; idxPoint < rayNbPoints; idxPoint ++) {
        printf ("point %d [%d %d], a= %d\n", idxPoint, lPointsX[idxPoint], lPointsY[idxPoint], lAngle[idxPoint]);
    }
}
void textZBuffer () {
    unsigned char ii, jj;
    for (jj=0; jj< 10; jj++)  {
        printf ("%d|%d|%d|%d \t %d|%d|%d|%d \t%d|%d|%d|%d\n", (jj), raywall[(jj)], rayzbuffer[(jj)], TableVerticalPos[jj], (jj)+10, raywall[(jj)+10], rayzbuffer[(jj)+10], TableVerticalPos[jj+10], (jj)+20, raywall[(jj)+20], rayzbuffer[(jj)+20] , TableVerticalPos[jj+20]);
    }
    for (jj=30; jj< 40; jj++)  {
        printf ("%d|%d|%d|%d\n", (jj), raywall[(jj)], rayzbuffer[(jj)], TableVerticalPos[jj]);
     }
}
#endif
