
#define max(a,b)            (((a) > (b)) ? (a) : (b))
#define min(a,b)            (((a) < (b)) ? (a) : (b))
#define abs(x)                 (((x)<0)?-(x):(x))


#ifdef USE_C_COSSIN
#define sin(x) tab_sin[(unsigned char)(x)]
#define cos(x) tab_cos[(unsigned char)(x)]
#endif // USE_C_COSSIN

#ifdef USE_C_LOG2
#define log2(x) log2_tab[(unsigned char)(x)]
#endif // USE_C_LOG2

#ifdef USE_C_ATAN2

signed char batan2 (signed int ty, signed int tx){
    signed char     x, y;
    unsigned char   ix, iy;
    unsigned char   res_div;
    unsigned char   idx;
    unsigned char   noct;       // Octant number
    unsigned char   v;

    if ((abs(tx)>127) || (abs(ty)>127)) {
        x=tx/2;
        y=ty/2;
    }else{
        x=tx;
        y=ty;
    }
    noct = 0;
    if (x<0){
        ix = -x ; // (unsigned char)(x ^ 0xFF);
        noct |= 4;
    } else { 
        ix = (unsigned char)x;
    }
    if (y<0){
        iy = -y; // (unsigned char)(y ^ 0xFF);
        noct |= 2;
    }else{
        iy = (unsigned char)y;
    }
    res_div = (log2_tab[ix]) - (log2_tab [iy]);
    // printf ("ix = %d iy = %d; res_div = %d\n", ix, iy, res_div);
                  
    if (log2_tab[ix] >= log2_tab [iy]){
        idx = res_div ^ 0xFF;
        noct |= 1;

    } else {
        idx = res_div;
    } 

    v= atan_tab[idx];
    v ^= octant_adjust[noct];

    return (signed char)v;
}
#endif // USE_C_ATAN2

#ifdef USE_ASM_ATAN2
extern signed char tx, ty, res;
signed char batan2 (signed int dY, signed int dX){
//void myAtan (signed char dY, signed char dX, signed char * angle){
    ty = dY;
    tx = dX;
    atan2_8();
    return (res);
}
#endif // USE_ASM_ATAN2

#ifdef USE_C_NORM
unsigned int norm (signed int dx, signed int dy) {
    unsigned char ax, ay, x, y;
    unsigned char dividedBy2;
    unsigned int N;

    if ((abs(dx)>=128) || (abs(dy)>=128)){
        ax= abs(dx)/2; ay = abs(dy)/2;
        dividedBy2 = 1;
    } else {
        ax= abs(dx); ay = abs(dy);
        dividedBy2 = 0;
    }

    if (ax >= ay){
        x = ax; y = ay;
    } else{
        x = ay; y = ax;
    }
    if (y > x/2){
        // N = (math.sqrt(5)-math.sqrt(2))*x + (2*math.sqrt(2) - math.sqrt(5))*y
        //N = tabmult_sqrt5_m_sqrt2 [x] + tabmult_2sqrt2_m_sqrt5[y]
        N = tabmult_C [x] + tabmult_D[y];
    } else {
        // N = x+(math.sqrt(5)/2 - 1)*y
        // N = x + tabmult_sqrt5_m2 [y]
        N = tabmult_A [x] + tabmult_B[y];
    }
    if (dividedBy2 != 0){
        return 2*N;
    } else{
        return N;
    }
}
#endif

#ifdef USE_ASM_NORM
extern signed char DeltaX, DeltaY;
extern unsigned char Norm;
extern void norm_8();
void norm (signed char dX, signed char dY, unsigned char * norm){
    DeltaY = dY;
    DeltaX = dX;
    norm_8();
    *norm = Norm;
}

#endif // USE_ASM_NORM

#ifdef USE_C_EXP
unsigned int LONGEXP(unsigned int x){
    if (x < 256) {        // 0000001 00000000
        return ((unsigned int)tab_exp[(unsigned char)(x)]);
    } else if (x < 289) { // 0000001 00100001
        return ((unsigned int)tab_exp_extended[(unsigned char)(x-256)]) + 256; // 00000001 00000000
    } else if (x < 307) { // 0000001 00110011
        return ((unsigned int)tab_exp_extended2[(unsigned char)(x-289)]) + 512; // 00000010 00000000
    } else if (x < 320) { // 0000001 01000000
        return ((unsigned int)tab_exp_extended3[(unsigned char)(x-307)]) + 756; // 00000010 11110100
    } else {
        return 0x7FFF;
    }
}
#define exp(x) LONGEXP((unsigned int)(x))
#endif // USE_C_EXP

#ifdef USE_ASM_EXP
unsigned int RayLogNumerator;
unsigned int RayAXoverB;
unsigned int LONGEXP(unsigned int x){
    RayLogNumerator = x;
    asm (
        "debug_here;"
        "lda #0;"
        // "sta _RayAXoverB;"
        "sta _RayAXoverB+1;"
        "lda _RayLogNumerator+1;"
        "beq RayLogNumeratorUnder256;"
        "lda _RayLogNumerator;"
        "cmp #$21;"
        "bmi RayLogNumeratorUnder289;"
        "cmp #$33;"
        "bmi RayLogNumeratorUnder307;"
        "cmp #$40;"
        "bmi RayLogNumeratorUnder320;"
    );asm(
            "lda #$FF;"
            "sta _RayAXoverB;"
            "lda #$7F;"
            "sta _RayAXoverB+1;"
        "jmp exponentiateRatio_done;"
        "RayLogNumeratorUnder320;"
// RayAXoverB= ((unsigned int)tab_exp_extended3[(unsigned char)(RayLogNumerator-307)]) + 756; // 00000010 11110100
            "sec;"
            "sbc #$33;" // LOBYT(307) = 00110011
            "tay;"
            "lda _tab_exp_extended3, y;"
            "clc;"
            "adc #$f4;"
            "sta _RayAXoverB;"
            "lda #$02;"
            "adc #0;"
            "sta _RayAXoverB+1;"
        "jmp exponentiateRatio_done;"
    );asm(
        "RayLogNumeratorUnder307;"
//RayAXoverB= ((unsigned int)tab_exp_extended2[(unsigned char)(RayLogNumerator-289)]) + 512; // 00000010 00000000
            "sec;"
            "sbc #$21;" // LOBYT(289) = 00100001
            "tay;"
            "lda _tab_exp_extended2, y;"
            "sta _RayAXoverB;"
            "lda #2;"  // HIBYT(512) = 00000010
            "sta _RayAXoverB+1;"

        "jmp exponentiateRatio_done;"
    );asm(
        "RayLogNumeratorUnder289;"
//RayAXoverB= ((unsigned int)tab_exp_extended[(unsigned char)(RayLogNumerator-256)]) + 256; // 00000001 00000000
            "sec;"
            "sbc #$00;" // LOBYT(256) = 00000000
            "tay;"
            "lda _tab_exp_extended, y;"
            "sta _RayAXoverB;"
            "lda #1;"
            "sta _RayAXoverB+1;"
        "jmp exponentiateRatio_done;"
    );asm(
        "RayLogNumeratorUnder256;"
            "ldy _RayLogNumerator;"
            "lda _tab_exp, y;"
            "sta _RayAXoverB;"
        "exponentiateRatio_done;"
    );    
    return (RayAXoverB);
}
#define exp(x) LONGEXP((unsigned int)(x))
#endif // USE_ASM_EXP
