#include "ray_param.h"

#define FOV_HALF_ANGLE                  SCREEN_WIDTH
#define SLICE_ANGULAR_INCREMENT         2
#define ANGLE_MAX                       (signed char)0xC0

#define IS_FRONT(i)                     (((lAngle[(i)] & ANGLE_MAX) == 0) || ((lAngle[(i)] & ANGLE_MAX) == ANGLE_MAX))

#ifdef USE_C_RAYCASTING_DATA
static signed char rayCamPosY, rayCamPosX, rayCamRotZ;
static unsigned char rayNbPoints;
static unsigned char rayNbWalls;

static signed char    lPointsX[NB_MAX_POINT];
static signed char    lPointsY[NB_MAX_POINT];
static unsigned char  lWallsPt1[NB_MAX_WALL];
static unsigned char  lWallsPt2[NB_MAX_WALL];
static unsigned char  rayzbuffer[NB_SLICES];
static unsigned char  TableVerticalPos[NB_SLICES];
static unsigned char  raywall[NB_SLICES];


static signed char    lAngle[NB_MAX_POINT];
static unsigned char  isFront[NB_MAX_POINT];
static unsigned char  isVisible[NB_MAX_POINT];
static signed char    lWallsCosBeta[NB_MAX_WALL];


static signed char    RayAlpha;
static signed char    RayLeftAlpha;
static signed char    RayRightAlpha;

static unsigned char  RayCurrentWall;
static unsigned char  RayIdXPoint1, RayIdXPoint2;
static unsigned char  RaySliceIdx;
static unsigned int   RayDistance;
static signed char    InterpAngleLeft;
static signed char    InterpAngleRight;
static unsigned char  InterpIdxLeft;
static unsigned char  RayNbSlice;
static unsigned char* tab_denom;
static unsigned int   RayWallLog;
#else
#ifdef USE_ASM_RAYCASTING_DATA
extern signed char rayCamPosY, rayCamPosX, rayCamRotZ;
extern unsigned char rayNbPoints;
extern unsigned char rayNbWalls;
extern signed char    lPointsX[NB_MAX_POINT];
extern signed char    lPointsY[NB_MAX_POINT];
extern unsigned char  lWallsPt1[NB_MAX_WALL];
extern unsigned char  lWallsPt2[NB_MAX_WALL];
extern unsigned char  rayzbuffer[NB_SLICES];
extern unsigned char  TableVerticalPos[NB_SLICES];
extern unsigned char  raywall[NB_SLICES];
extern signed char    lAngle[NB_MAX_POINT];
extern unsigned char  isFront[NB_MAX_POINT];
extern unsigned char  isVisible[NB_MAX_POINT];
extern signed char    lWallsCosBeta[NB_MAX_WALL]; 
extern signed char    RayAlpha;
extern signed char    RayLeftAlpha;
extern signed char    RayRightAlpha; 
extern unsigned char  RayCurrentWall;
extern unsigned char  RayIdXPoint1, RayIdXPoint2;
extern unsigned char  RaySliceIdx;
extern unsigned int   RayDistance;
extern signed char    InterpAngleLeft;
extern signed char    InterpAngleRight;
extern unsigned char  InterpIdxLeft;
extern unsigned char  RayNbSlice;
extern unsigned char* tab_denom;
extern unsigned int   RayWallLog;
#endif // USE_ASM_RAYCASTING_DATA
#endif // USE_C_RAYCASTING_DATA

void initScene (signed char sceneData[]);
void rayInitCasting();
unsigned char dist2hh(unsigned int x);
static void rayProcessPoints();




#ifdef USE_C_RAYCASTING
static void drawFullVisibleWall();

static void drawFullCrossingWall();
static void drawLeftCuttingWall1Visible();
static void drawLeftCuttingWall2Visible();
static void drawRightCuttingWall1Visible();
static void drawRightCuttingWall2Visible();
#endif // USE_C_RAYCASTING

void rayInitCasting(){
    unsigned char ii;
    for (ii=0; ii< NB_SLICES; ii++) {
        rayzbuffer[ii]      = 255;
        raywall[ii]         = 255;
    }
    // memset (rayzbuffer, 0xFF, NB_SLICES);
    // memset (raywall, 255, NB_SLICES);
}

static void rayProcessPoints() {

    for (RayIdXPoint1 = 0; RayIdXPoint1 < rayNbPoints; RayIdXPoint1 ++) {

        lAngle[RayIdXPoint1] =  batan2(lPointsY[RayIdXPoint1]-rayCamPosY, lPointsX[RayIdXPoint1]-rayCamPosX) - rayCamRotZ;

        isVisible[RayIdXPoint1]         = (abs(lAngle[RayIdXPoint1])<FOV_HALF_ANGLE)?1:0;
        isFront[RayIdXPoint1]           = IS_FRONT(RayIdXPoint1)?1:0;
    }
}
void rayProcessWalls() {

    for (RayCurrentWall = 0; RayCurrentWall < rayNbWalls; RayCurrentWall ++){
        RayIdXPoint1        = lWallsPt1[RayCurrentWall];
        RayIdXPoint2        = lWallsPt2[RayCurrentWall];

        if (isVisible[RayIdXPoint1]) {
            if (isVisible[RayIdXPoint2]) {
                drawFullVisibleWall();
            } else {
                // Point 1 is visible and Point 2 is not in FOV
                if (lAngle[RayIdXPoint1] > 0){
                    if ((lAngle[RayIdXPoint1] & 0x80) == (lAngle[RayIdXPoint2] & 0x80)) {
                        drawLeftCuttingWall1Visible();
                    } else {
#ifdef USE_C_RAYCASTING
                        if (abs(lAngle[RayIdXPoint1] - lAngle[RayIdXPoint2]) > 127){
                            drawLeftCuttingWall1Visible();
                        } else {
                            drawRightCuttingWall1Visible();
                        }
#else 
                        asm (
                            "debugHere;"
                            "ldy _RayIdXPoint2;"
                            "lda _lAngle,Y;"
                            "sta reg7;"
                            "ldy _RayIdXPoint1;"
                            "lda _lAngle,Y;"
                            "sec;"
                            "sbc reg7;"
                            "bvc VeryWideAngle_01;"
                            "jsr _drawLeftCuttingWall1Visible;"
                            "jmp endAngleSpecific_01;"
                        "VeryWideAngle_01;"
                            "jsr _drawRightCuttingWall1Visible;"
                        "endAngleSpecific_01;"

                        );
#endif 
                    }

                } else if (lAngle[RayIdXPoint1] < 0) {
                    if ((lAngle[RayIdXPoint1] & 0x80) == (lAngle[RayIdXPoint2] & 0x80)) {
                        drawRightCuttingWall1Visible();
                    } else {
#ifdef USE_C_RAYCASTING
                        if (abs(lAngle[RayIdXPoint1] - lAngle[RayIdXPoint2]) > 127){
                            drawRightCuttingWall1Visible();
                        } else {
                            drawLeftCuttingWall1Visible();
                        }
#else
                        asm (
                            "ldy _RayIdXPoint2;"
                            "lda _lAngle,Y;"
                            "sta reg7;"
                            "ldy _RayIdXPoint1;"
                            "lda _lAngle,Y;"
                            "sec;"
                            "sbc reg7;"
                            "bvc VeryWideAngle_02;"
                            "jsr _drawRightCuttingWall1Visible;"
                            "jmp endAngleSpecific_02;"
                        "VeryWideAngle_02;"
                            "jsr _drawLeftCuttingWall1Visible;"
                        "endAngleSpecific_02;"

                        );
#endif
                    }

                } else { // (lAngle[idxPt1] < 0)
#ifdef USE_C_RAYCASTING
                    if (lAngle[RayIdXPoint2] < 0) {
                        drawRightCuttingWall1Visible();
                    } else {
                        drawLeftCuttingWall1Visible();
                    }
#else
                    asm (
                        "ldy _RayIdXPoint2;"
                        "lda _lAngle,Y;"
                        "bpl PositivPoint2Angle;"
                            "jsr _drawRightCuttingWall1Visible;"
                        "jmp endPositivPoint2Angle;"
                        "PositivPoint2Angle;"
                            "jsr _drawLeftCuttingWall1Visible;"
                        "endPositivPoint2Angle;"

                    );
#endif
                }
            }
        } else {
            if (isVisible[RayIdXPoint2]) {

                // Point 2 is visible and Point 1 is not in FOV
                if (lAngle[RayIdXPoint2] > 0){
                    if ((lAngle[RayIdXPoint1] & 0x80) == (lAngle[RayIdXPoint2] & 0x80)) {
                        drawLeftCuttingWall2Visible();
                    } else {
#ifdef USE_C_RAYCASTING
                        if (abs(lAngle[RayIdXPoint2] - lAngle[RayIdXPoint1]) > 127){
                            drawLeftCuttingWall2Visible();
                        } else {
                            drawRightCuttingWall2Visible();
                        }
#else
                        asm (
                            "ldy _RayIdXPoint1;"
                            "lda _lAngle,Y;"
                            "sta reg7;"
                            "ldy _RayIdXPoint2;"
                            "lda _lAngle,Y;"
                            "sec;"
                            "sbc reg7;"
                            "bvc VeryWideAngle_03;"
                            "jsr _drawLeftCuttingWall2Visible;"
                            "jmp endAngleSpecific_03;"
                        "VeryWideAngle_03;"
                            "jsr _drawRightCuttingWall2Visible;"
                        "endAngleSpecific_03;"

                        );
#endif
                    }
                } else if (lAngle[RayIdXPoint2] < 0) {
                    if ((lAngle[RayIdXPoint1] & 0x80) == (lAngle[RayIdXPoint2] & 0x80)) {
                        drawRightCuttingWall2Visible();
                    } else {
#ifdef USE_C_RAYCASTING

                        if (abs(lAngle[RayIdXPoint2] - lAngle[RayIdXPoint1]) > 127){
                            drawRightCuttingWall2Visible();
                        } else {
                            drawLeftCuttingWall2Visible();
                        }
#else
                        asm (
                            "ldy _RayIdXPoint1;"
                            "lda _lAngle,Y;"
                            "sta reg7;"
                            "ldy _RayIdXPoint2;"
                            "lda _lAngle,Y;"
                            "sec;"
                            "sbc reg7;"
                            "bvc VeryWideAngle_04;"
                            "jsr _drawRightCuttingWall2Visible;"
                            "jmp endAngleSpecific_04;"
                        "VeryWideAngle_04;"
                            "jsr _drawLeftCuttingWall2Visible;"
                        "endAngleSpecific_04;"

                        );
#endif
                    }
                } else { // (lAngle[idxPt1] == 0
#ifdef USE_C_RAYCASTING
                    if (lAngle[RayIdXPoint1] < 0) {
                        drawRightCuttingWall2Visible();
                    } else {
                        drawLeftCuttingWall2Visible();
                    }
#else 
                    asm (
                        "ldy _RayIdXPoint1;"
                        "lda _lAngle,Y;"
                        "bpl PositivPoint2Angle_02;"
                            "jsr _drawRightCuttingWall2Visible;"
                        "jmp endPositivPoint2Angle_02;"
                        "PositivPoint2Angle_02;"
                            "jsr _drawLeftCuttingWall2Visible;"
                        "endPositivPoint2Angle_02;"

                    );
#endif
                }
            } else {
                // asm("debugHere;");
                if (   (isFront[RayIdXPoint1] 
                    || isFront[RayIdXPoint2] )
                    && ((lAngle[RayIdXPoint1] & 0x80) != (lAngle[RayIdXPoint2] & 0x80)) 
                    && (abs(lAngle[RayIdXPoint2]) < 127 - abs(lAngle[RayIdXPoint1])) 
                ) {
                    drawFullCrossingWall();
                }
            }
        }
    }
    /* 
     * Change output from logarithmic scale to linear scale 
     */
    for (RaySliceIdx=0; RaySliceIdx<NB_SLICES; RaySliceIdx++){

#ifdef USE_ANTIFISH
        rayzbuffer[RaySliceIdx] -= unfish[RaySliceIdx];
#endif
#ifdef USE_DISTANCE2HIRESHEIGHT
        TableVerticalPos[RaySliceIdx] =dist2hh(rayzbuffer[RaySliceIdx]);
        rayzbuffer[RaySliceIdx] = exp(rayzbuffer[RaySliceIdx]);        
#endif
    }
}

static void precalculateWallsAngle();

void initScene (signed char sceneData[]){
	unsigned int ii;
	unsigned char jj;

	ii=0;
	rayNbPoints = (unsigned char)sceneData[ii++];
	rayNbWalls = (unsigned char)sceneData[ii++];
	for (jj=0; jj < rayNbPoints; jj++){
		lPointsX[jj]= sceneData[ii++] ; lPointsY[jj] = sceneData[ii++];  // points 0
	}
	for (jj=0; jj < rayNbWalls; jj++){
		lWallsPt1[jj]= (unsigned char)(sceneData[ii++]) ; lWallsPt2[jj] = (unsigned char)(sceneData[ii++]);// points 0
	}
    precalculateWallsAngle();
}


static void precalculateWallsAngle() {
    unsigned char idxWall, idxPt1, idxPt2;
    signed char dX, dY, angle;

    for (idxWall = 0; idxWall < rayNbWalls; idxWall ++) { 

        idxPt1 = lWallsPt1[idxWall];
        idxPt2 = lWallsPt2[idxWall];

        dX = lPointsX[idxPt2]-lPointsX[idxPt1];
        dY = lPointsY[idxPt2]-lPointsY[idxPt1];

        if (dX == 0) {
            lWallsCosBeta[idxWall] = 0;
        } else if (dY == 0) {
            if (dX > 0) {
                lWallsCosBeta[idxWall] = 32;
            } else {
                lWallsCosBeta[idxWall] = -32;
            }
        } else {
            /* 
             *  Not aligned walls not handled
             */
        }
    }
}

unsigned char dist2hh(unsigned int x){
    if (x < 256) {
        return ((unsigned char)unlogd2hh[(unsigned char)(x)]);
    } else return 100;
}


#ifdef USE_C_TOTO
void toto() {

    RayAlpha        = InterpAngleLeft;
    RaySliceIdx     = InterpIdxLeft;
    if (RayNbSlice == 0) return;

#ifdef DEBUG
    printf ("Wall %d from %d to %d, %d slice from (c=%d a=%d) to (c=%d, a=%d) \n", RayCurrentWall, RayIdXPoint1, RayIdXPoint2,
    RayNbSlice, RaySliceIdx,RayAlpha, RaySliceIdx+RayNbSlice, RayAlpha-RayNbSlice*SLICE_ANGULAR_INCREMENT
    );
#endif
    
    while(1){
        // RayDistance = longexp(RayDistance);
        RayDistance = RayWallLog + (unsigned int)tab_denom[(unsigned char)RayAlpha];

#ifdef DEBUG
        printf ("RaySliceIdx = %d, RayAlpha = %d, RayDistance=%d\n", RaySliceIdx, RayAlpha, RayDistance);
#endif
        if (rayzbuffer[RaySliceIdx] > RayDistance){
            rayzbuffer[RaySliceIdx] = RayDistance;
            raywall[RaySliceIdx] = RayCurrentWall;        
        }

        RaySliceIdx += 1;

        RayAlpha-=SLICE_ANGULAR_INCREMENT;

        if ((--RayNbSlice) == 0) break;
    }
}
#endif // USE_C_TOTO

#ifdef USE_C_PREDRAW
static void preDraw(){

    if (lWallsCosBeta[RayCurrentWall] == 0){    // Wall is O,y aligned   
        RayWallLog = log2_tab[(unsigned char)(abs(lPointsX[RayIdXPoint1]-rayCamPosX))];
        // asm (
        //     "ldy _RayIdXPoint1;"
        //     "lda _lPointsX,Y;"
        //     "sec;"
        //     "sbc _rayCamPosX;"
        //     "bpl PositivOyDeltaWithWall;"
        //     "eor #$FF;"
        //     "clc;"
        //     "adc #1;"
        //     "PositivOyDeltaWithWall;"
        //     "tay;"
        //     "lda _log2_tab,Y;"
        //     "sta _RayWallLog;"

        // );
        tab_denom = tab_1overcos;
        // asm (
        // "lda #<(_tab_1overcos);"
        // "sta _tab_denom;"
        // "lda #>(_tab_1overcos);"
        // "sta _tab_denom+1;"
        // );
    } else {                                    // Wall is O,x aligned
        RayWallLog = log2_tab[(unsigned char)(abs(lPointsY[RayIdXPoint1]-rayCamPosY))];
        // asm (
        //     "ldy _RayIdXPoint1;"
        //     "lda _lPointsY,Y;"
        //     "sec;"
        //     "sbc _rayCamPosY;"
        //     "bpl PositivOxDeltaWithWall;"
        //     "eor #$FF;"
        //     "clc;"
        //     "adc #1;"
        //     "PositivOxDeltaWithWall;"
        //     "tay;"
        //     "lda _log2_tab,Y;"
        //     "sta _RayWallLog;"

        // );
        tab_denom = tab_1oversin;
        // asm (
        // "lda #<(_tab_1oversin);"
        // "sta _tab_denom;"
        // "lda #>(_tab_1oversin);"
        // "sta _tab_denom+1;"
        // );
    }

}
#endif // USE_C_PREDRAW




#ifdef USE_C_RAYCASTING



static void drawFullVisibleWall(){

    preDraw();

    if (lAngle[RayIdXPoint1] > lAngle[RayIdXPoint2]){
        InterpIdxLeft = (SCREEN_WIDTH - lAngle[RayIdXPoint1]) >> 1;
        InterpAngleLeft=lAngle[RayIdXPoint1]+ rayCamRotZ;
        InterpAngleRight=lAngle[RayIdXPoint2]+ rayCamRotZ; // TODO : remove me to save cycles
        RayNbSlice = ((lAngle[RayIdXPoint1] - lAngle[RayIdXPoint2]) >> 1 )+1;
    } else {
        InterpIdxLeft = (SCREEN_WIDTH - lAngle[RayIdXPoint2]) >> 1;
        InterpAngleLeft=lAngle[RayIdXPoint2]+ rayCamRotZ;
        InterpAngleRight=lAngle[RayIdXPoint1]+ rayCamRotZ; // TODO : remove me to save cycles
        RayNbSlice = ((lAngle[RayIdXPoint2] - lAngle[RayIdXPoint1]) >> 1 )+1;
    }

    toto();
}

static void drawFullCrossingWall(){

    preDraw();

    InterpAngleLeft     =RayLeftAlpha;
    InterpAngleRight    =RayRightAlpha;

    RayNbSlice          = NB_SLICES;
    InterpIdxLeft       = 0;

    toto();
}

static void drawLeftCuttingWall1Visible(){
    
    preDraw();
    InterpAngleLeft=RayLeftAlpha;
    InterpAngleRight=lAngle[RayIdXPoint1]+ rayCamRotZ;
    InterpIdxLeft = 0; 
    RayNbSlice = (((InterpAngleLeft - InterpAngleRight)&0xFF) >> 1)+1;
    toto();

    // asm (
    //     "jsr _preDraw;"
    // // InterpAngleLeft=RayLeftAlpha;   
    //     "lda _RayLeftAlpha;"
    //     "sta _InterpAngleLeft;"

    // // InterpAngleRight=lAngle[RayIdXPoint1]+ rayCamRotZ;
    //     "ldy _RayIdXPoint1;"
    //     "lda _lAngle,Y;"
    //     "clc;"
    //     "adc _rayCamRotZ;"
    //     "sta _InterpAngleRight;"

    // // InterpIdxLeft = 0; 
    //     "lda #0;"
    //     "sta _InterpIdxLeft;"

    // // RayNbSlice = (((InterpAngleLeft - InterpAngleRight)&0xFF) >> 1)+1;
    //     "lda _InterpAngleLeft;"
    //     "sec;"
    //     "sbc _InterpAngleRight;"
    //     "lsr;"
    //     "clc;"
    //     "adc #1;"
    //     "sta _RayNbSlice;"

    //     "jsr _toto;"
    // );
}

static void drawLeftCuttingWall2Visible(){
    preDraw();
    InterpAngleLeft=RayLeftAlpha;
    InterpAngleRight=lAngle[RayIdXPoint2]+ rayCamRotZ;
    InterpIdxLeft = 0; 
    RayNbSlice = (((InterpAngleLeft - InterpAngleRight)&0xFF) >> 1)+1;
    toto();

    // asm (
    //     "jsr _preDraw;"

    // // InterpAngleLeft=RayLeftAlpha;
    //     "lda _RayLeftAlpha;"
    //     "sta _InterpAngleLeft;"

    // // InterpAngleRight=lAngle[RayIdXPoint2]+ rayCamRotZ;
    //     "ldy _RayIdXPoint2;"
    //     "lda _lAngle,Y;"
    //     "clc;"
    //     "adc _rayCamRotZ;"
    //     "sta _InterpAngleRight;"

    // // InterpIdxLeft = 0; 
    //     "lda #0;"
    //     "sta _InterpIdxLeft;"

    // // RayNbSlice = (((InterpAngleLeft - InterpAngleRight)&0xFF) >> 1)+1;
    //     "lda _InterpAngleLeft;"
    //     "sec;"
    //     "sbc _InterpAngleRight;"
    //     "lsr;"
    //     "adc #1;"
    //     "sta _RayNbSlice;"

    //     "jsr _toto;"
    // );

}

static void drawRightCuttingWall1Visible(){
    preDraw();
    InterpIdxLeft = (SCREEN_WIDTH - lAngle[RayIdXPoint1]) >> 1;
    InterpAngleLeft=lAngle[RayIdXPoint1]+rayCamRotZ;
    InterpAngleRight=RayRightAlpha;
    RayNbSlice = (((InterpAngleLeft - InterpAngleRight)&0xFF) >> 1)+1;
    toto();
    
    // asm (
    //     "jsr _preDraw;"
    //     // InterpIdxLeft = (SCREEN_WIDTH - lAngle[RayIdXPoint1]) >> 1;
    //     "ldy _RayIdXPoint1;"
    //     "lda #40;"
    //     "sec;"
    //     "sbc _lAngle,Y;"
    //     "lsr;"
    //     "sta _InterpIdxLeft;"

    //     // InterpAngleLeft=lAngle[RayIdXPoint1]+rayCamRotZ;
    //     "lda _rayCamRotZ;"
    //     "ldy _RayIdXPoint1;"
    //     "clc;"
    //     "adc _lAngle,Y;"
    //     "sta _InterpAngleLeft;"

    //     // InterpAngleRight=RayRightAlpha;
    //     "lda _RayRightAlpha;"
    //     "sta _InterpAngleRight;"

    //     // RayNbSlice = (((InterpAngleLeft - InterpAngleRight)&0xFF) >> 1)+1;
    //     "lda _InterpAngleLeft;"
    //     "sec;"
    //     "sbc _InterpAngleRight;"
    //     "lsr;"
    //     "adc #1;"
    //     "sta _RayNbSlice;"
    //     "jsr _toto;"
    // );



}
static void drawRightCuttingWall2Visible(){
    preDraw();
    InterpAngleLeft=lAngle[RayIdXPoint2]+rayCamRotZ;
    InterpAngleRight=RayRightAlpha;
    InterpIdxLeft = (SCREEN_WIDTH - lAngle[RayIdXPoint2]) >> 1;
    RayNbSlice = (((InterpAngleLeft - InterpAngleRight)&0xFF) >> 1)+1;
    toto();

    // asm (
    //     "jsr _preDraw;"

    // // InterpAngleLeft=lAngle[RayIdXPoint2]+rayCamRotZ;
    //     "lda _rayCamRotZ;"
    //     "ldy _RayIdXPoint2;"
    //     "clc;"
    //     "adc _lAngle,Y;"
    //     "sta _InterpAngleLeft;"

    // // InterpAngleRight=RayRightAlpha;
    //     "lda _RayRightAlpha;"
    //     "sta _InterpAngleRight;"

    // // InterpIdxLeft = (SCREEN_WIDTH - lAngle[RayIdXPoint2]) >> 1;
    //     "ldy _RayIdXPoint2;"
    //     "lda #40;"
    //     "sec;"
    //     "sbc _lAngle,Y;"
    //     "lsr;"
    //     "sta _InterpIdxLeft;"

    // // RayNbSlice = (((InterpAngleLeft - InterpAngleRight)&0xFF) >> 1)+1;
    //     "lda _InterpAngleLeft;"
    //     "sec;"
    //     "sbc _InterpAngleRight;"
    //     "lsr;"
    //     "adc #1;"
    //     "sta _RayNbSlice;"


    //     "jsr _toto;"
    // );

}
#endif // USE_C_RAYCASTING